@@ -37,6 +37,7 @@
 #include <netdb.h>
 #include <net/if.h>
 #include <netinet/tcp.h>
+#include <ifaddrs.h>
 
 #include "giognutls.h"
 #include "gresolv.h"
@@ -97,6 +98,7 @@ struct web_session {
 	GWebResult result;
 
 	GWebResultFunc result_func;
+	GWebRouteFunc route_func;
 	GWebInputFunc input_func;
 	int fd;
 	gsize length;
@@ -441,6 +443,13 @@ static inline void call_result_func(struct web_session *session, guint16 status)
 					result == TRUE ? "continue" : "stop");
 }
 
+static inline void call_route_func(struct web_session *session)
+{
+	if (session->route_func != NULL)
+		session->route_func(session->address, session->addr->ai_family,
+				session->web->index, session->user_data);
+}
+
 static gboolean process_send_buffer(struct web_session *session)
 {
 	GString *buf;
@@ -954,6 +963,57 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 	return TRUE;
 }
 
+static int bind_to_address(int sk, const char *interface, int family)
+{
+	struct ifaddrs *ifaddr_list, *ifaddr;
+	int size, err = -1;
+
+	if (getifaddrs(&ifaddr_list) < 0)
+		return err;
+
+	for (ifaddr = ifaddr_list; ifaddr != NULL; ifaddr = ifaddr->ifa_next) {
+		if (g_strcmp0(ifaddr->ifa_name, interface) != 0)
+			continue;
+
+		if (ifaddr->ifa_addr == NULL ||
+				ifaddr->ifa_addr->sa_family != family)
+			continue;
+
+		switch (family) {
+		case AF_INET:
+			size = sizeof(struct sockaddr_in);
+			break;
+		case AF_INET6:
+			size = sizeof(struct sockaddr_in6);
+			break;
+		default:
+			continue;
+		}
+
+		err = bind(sk, (struct sockaddr *) ifaddr->ifa_addr, size);
+		break;
+	}
+
+	freeifaddrs(ifaddr_list);
+	return err;
+}
+
+static inline int bind_socket(int sk, int index, int family)
+{
+	char interface[IF_NAMESIZE];
+	int err;
+
+	if (if_indextoname(index, interface) == NULL)
+		return -1;
+
+	err = setsockopt(sk, SOL_SOCKET, SO_BINDTODEVICE,
+					interface, IF_NAMESIZE);
+	if (err < 0)
+		err = bind_to_address(sk, interface, family);
+
+	return err;
+}
+
 static int connect_session_transport(struct web_session *session)
 {
 	GIOFlags flags;
@@ -965,18 +1025,11 @@ static int connect_session_transport(struct web_session *session)
 		return -EIO;
 
 	if (session->web->index > 0) {
-		char interface[IF_NAMESIZE];
-
-		memset(interface, 0, IF_NAMESIZE);
-
-		if (if_indextoname(session->web->index, interface) != NULL) {
-			if (setsockopt(sk, SOL_SOCKET, SO_BINDTODEVICE,
-						interface, IF_NAMESIZE) < 0) {
-				close(sk);
-				return -EIO;
-			}
-
-			debug(session->web, "Use interface %s", interface);
+		if (bind_socket(sk, session->web->index,
+					session->addr->ai_family) < 0) {
+			debug(session->web, "bind() %s", strerror(errno));
+			close(sk);
+			return -EIO;
 		}
 	}
 
@@ -989,6 +1042,7 @@ static int connect_session_transport(struct web_session *session)
 	}
 
 	if (session->transport_channel == NULL) {
+		debug(session->web, "channel missing");
 		close(sk);
 		return -ENOMEM;
 	}
@@ -1005,6 +1059,7 @@ static int connect_session_transport(struct web_session *session)
 	if (connect(sk, session->addr->ai_addr,
 			session->addr->ai_addrlen) < 0) {
 		if (errno != EINPROGRESS) {
+			debug(session->web, "connect() %s", strerror(errno));
 			close(sk);
 			return -EIO;
 		}
@@ -1165,6 +1220,7 @@ static void resolv_result(GResolvResultStatus status,
 	}
 
 	session->address = g_strdup(results[0]);
+	call_route_func(session);
 
 	if (create_transport(session) < 0) {
 		call_result_func(session, 409);
@@ -1175,7 +1231,7 @@ static void resolv_result(GResolvResultStatus status,
 static guint do_request(GWeb *web, const char *url,
 				const char *type, GWebInputFunc input,
 				int fd, gsize length, GWebResultFunc func,
-				gpointer user_data)
+				GWebRouteFunc route, gpointer user_data)
 {
 	struct web_session *session;
 
@@ -1208,6 +1264,7 @@ static guint do_request(GWeb *web, const char *url,
 	session->web = web;
 
 	session->result_func = func;
+	session->route_func = route;
 	session->input_func = input;
 	session->fd = fd;
 	session->length = length;
@@ -1277,17 +1334,17 @@ static guint do_request(GWeb *web, const char *url,
 	return web->next_query_id++;
 }
 
-guint g_web_request_get(GWeb *web, const char *url,
-				GWebResultFunc func, gpointer user_data)
+guint g_web_request_get(GWeb *web, const char *url, GWebResultFunc func,
+		GWebRouteFunc route, gpointer user_data)
 {
-	return do_request(web, url, NULL, NULL, -1, 0, func, user_data);
+	return do_request(web, url, NULL, NULL, -1, 0, func, route, user_data);
 }
 
 guint g_web_request_post(GWeb *web, const char *url,
 				const char *type, GWebInputFunc input,
 				GWebResultFunc func, gpointer user_data)
 {
-	return do_request(web, url, type, input, -1, 0, func, user_data);
+	return do_request(web, url, type, input, -1, 0, func, NULL, user_data);
 }
 
 guint g_web_request_post_file(GWeb *web, const char *url,
@@ -1305,7 +1362,8 @@ guint g_web_request_post_file(GWeb *web, const char *url,
 	if (fd < 0)
 		return 0;
 
-	ret = do_request(web, url, type, NULL, fd, st.st_size, func, user_data);
+	ret = do_request(web, url, type, NULL, fd, st.st_size, func, NULL,
+			user_data);
 	if (ret == 0)
 		close(fd);
 
