@@ -55,8 +55,18 @@ struct connman_device;
 
 struct connman_device *connman_device_create(const char *node,
 						enum connman_device_type type);
-struct connman_device *connman_device_ref(struct connman_device *device);
-void connman_device_unref(struct connman_device *device);
+
+#define connman_device_ref(device) \
+	connman_device_ref_debug(device, __FILE__, __LINE__, __func__)
+
+#define connman_device_unref(device) \
+	connman_device_unref_debug(device, __FILE__, __LINE__, __func__)
+
+struct connman_device *
+connman_device_ref_debug(struct connman_device *device,
+			const char *file, int line, const char *caller);
+void connman_device_unref_debug(struct connman_device *device,
+			const char *file, int line, const char *caller);
 
 enum connman_device_type connman_device_get_type(struct connman_device *device);
 void connman_device_set_index(struct connman_device *device, int index);
@@ -72,12 +82,11 @@ int connman_device_set_powered(struct connman_device *device,
 						connman_bool_t powered);
 int connman_device_set_scanning(struct connman_device *device,
 						connman_bool_t scanning);
+void connman_device_reset_scanning(struct connman_device *device);
 
 int connman_device_set_disconnected(struct connman_device *device,
 						connman_bool_t disconnected);
 connman_bool_t connman_device_get_disconnected(struct connman_device *device);
-int connman_device_set_connected(struct connman_device *device,
-						connman_bool_t connected);
 
 int connman_device_set_string(struct connman_device *device,
 					const char *key, const char *value);
@@ -89,11 +98,9 @@ int connman_device_add_network(struct connman_device *device,
 struct connman_network *connman_device_get_network(struct connman_device *device,
 							const char *identifier);
 int connman_device_remove_network(struct connman_device *device,
-							const char *identifier);
+					struct connman_network *network);
 void connman_device_remove_all_networks(struct connman_device *device);
 
-void connman_device_schedule_scan(struct connman_device *device);
-
 int connman_device_register(struct connman_device *device);
 void connman_device_unregister(struct connman_device *device);
 
@@ -109,6 +116,10 @@ struct connman_device_driver {
 	int (*enable) (struct connman_device *device);
 	int (*disable) (struct connman_device *device);
 	int (*scan) (struct connman_device *device);
+	int (*scan_fast) (struct connman_device *device);
+	int (*scan_hidden)(struct connman_device *device,
+			const char *ssid, unsigned int ssid_len,
+			const char *identity, const char* passphrase);
 };
 
 int connman_device_driver_register(struct connman_device_driver *driver);
